#!/usr/bin/env python3

import pathlib
import pandas as pd
import datetime
import pickle
import sys

input_file = sys.argv[1]    #'./data/politicians/Ratsmitglieder_1848_DE_corr.xlsx'
output_file_csv = sys.argv[2]     #'./data/politicians/MPs_after1890.csv'
output_folder_dict = sys.argv[3]    


class MPs_Extractor(object):

    def __init__(self, years, input_file, output_file_csv, output_folder_dict, df_exc):
        self.input_file = input_file
        self.output_file_csv = output_file_csv
        self.output_folder_dict = output_folder_dict
        self.range_years = range(years[0], years[1] + 1)
        self.df_exc = df_exc
        
    # function to get lists of lastnames
    # input:
    # - df_year: dataframe for a year
    # output:
    # - list_names: 
    #      contains:
    #        - list of last names that appear only once and cannot be split
    #        - list of last name that are made up of two names such as 'Meier-Müller'
    #                    for each double name, four entries are made:
    #                        - ('Meier', 'Meier-Müller')
    #                        - ('Müller', 'Meier-Müller')
    #                        - ('Meier-Müller', 'Meier-Müller')
    #                        - ('MeierMüller', 'Meier-Müller')
    #        - list for composite last names such as 'von Arx' or 'de Stoppani'
    #                  will be saved as ('Arx', 'von Arx')
    #        - list for people with the same last names
    #                     will be saved as (lastname, lastname (canton)) for each person
    #                     if the name is a composite name: ('Arx', 'von Arx (canton)')
    def get_list_of_lastnames(self, df_year, df_after1890):
        str_simple = 'simple'
        str_double = 'double'
        str_comp = 'comp'
        str_canton2 = 'canton'

	# function to split lastname and save meaningful part(s) to list 
        def split_lastname(lastname, uniqueID, tpl_canton, str_canton = ''):
	    # if last name is a composite name, e.g. 'von Arx' and 'de Stoppani'
            lastname_split = lastname.split()
            if len(lastname_split) > 1:
                for item in lastname_split:
                    if item not in ['von', 'de', 'Ab', 'van']:
			# write distinctive item to extended list
                        if str_canton:
                            list_names.append((str_canton2, item, str_canton, uniqueID) + tpl_canton)
                        else:
                            list_names.append((str_comp, item, lastname, uniqueID) + tpl_canton)
            else:
		# if last name is a double name, e.g. 'Meier-Müller'
                lastname_split2 = lastname.replace('-', ' ').split()
                if len(lastname_split2) > 1:
		    # write each part of double name into corresponding list
                    for item in lastname_split2:
                        if str_canton:
                            list_names.append((str_canton2, item, str_canton, uniqueID) + tpl_canton)
                        else:
                            list_names.append((str_double, item, lastname, uniqueID) + tpl_canton)
		    # write double name into list
                    list_names.append((str_double, lastname, lastname, uniqueID) + tpl_canton)
		    # write double name without space into list
                    list_names.append((str_double, ''.join(lastname.split('-')), lastname, uniqueID) + tpl_canton)
                else:       
                    if str_canton:
                        list_names.append((str_canton2, lastname, str_canton, uniqueID) + tpl_canton)
                    else:
                        list_names.append((str_simple, lastname, lastname, uniqueID) + tpl_canton)

        # function to get canton and citizenship for uniqueID
        def get_canton(df_year, uniqueID):
            str_cantonname = df_year['CantonName'].loc[df_year['uniqueIndex']==uniqueID].iloc[0]
            str_cantonabbr = df_year['CantonAbbreviation'].loc[df_year['uniqueIndex']==uniqueID].iloc[0]
            str_citizenship = df_year['Citizenship'].loc[df_year['uniqueIndex']==uniqueID].iloc[0]
            str_firstname = df_year['FirstName'].loc[df_year['uniqueIndex']==uniqueID].iloc[0]
 
            return (str_cantonname, str_cantonabbr, str_citizenship, str_firstname)

	# create empty lists for last names
        list_names = []
	
	# for every last name
        for lastname in df_year['LastName'].drop_duplicates():
	    #print('name', lastname, type(lastname))
	    
	    # extract all entries with that last name
            df_temp = df_year.loc[df_after1890['LastName']==lastname]
	    #print(df_temp)
	    
	    # if there is an extra double name
            if df_temp.iloc[0]['DoubleName'] != '':
		# extract unique index
                uniqueID = df_temp.iloc[0]['uniqueIndex']
		
                # get canton information for that uniqueID
                tpl_canton = get_canton(df_year, uniqueID)

                #print('double name', df_temp)
                doublename = df_temp.iloc[0]['DoubleName']

		# if last name is a double name, e.g. 'Meier-Müller'
                lastname_split2 = doublename.replace('-', ' ').split()
                if len(lastname_split2) > 1:
		    # write each part of double name into corresponding list
                    for item in lastname_split2:
                        list_names.append((str_double, item, lastname, uniqueID) + tpl_canton)
		    # write double name into list
                    list_names.append((str_double, doublename, lastname, uniqueID) + tpl_canton)
		    # write double name without space into list
                    list_names.append((str_double, ''.join(doublename.split('-')), lastname, uniqueID) + tpl_canton)

	    # if only one person with that last name
            if df_temp.drop_duplicates(['uniqueIndex']).shape[0] == 1:
		# extract unique index
                uniqueID = df_temp.iloc[0]['uniqueIndex']

                # get canton information for that uniqueID
                tpl_canton = get_canton(df_year, uniqueID)

		# write complete name to list of last names
                split_lastname(lastname, uniqueID, tpl_canton)
		
	    # if there are several people with the same last name
            else:
		# write last name and canton to correct list
                for idx, row in df_temp.drop_duplicates(['uniqueIndex']).iterrows():
		    # extract unique index
                    uniqueID = df_temp.loc[idx]['uniqueIndex']
		    
                    # get canton information for that uniqueID
                    tpl_canton = get_canton(df_year, uniqueID)

		    # write the lastname to the list
                    split_lastname(lastname, uniqueID, tpl_canton, row['LastName'] + ' (' + row['CantonName'] + ')') 
			
        return list_names

    def extract(self):
        # read excel file and save first sheet to a dataframe
        xl = pd.ExcelFile(self.input_file)
        str_sheetname = xl.sheet_names[0]
        orddict = xl.parse([str_sheetname])
        df = orddict[str_sheetname]

        # drop duplicate entries
        df = df.drop_duplicates()

        # extract all people participating after 1890
        # starting from 1891
        df1 = df[pd.to_datetime(df['DateLeaving']) > datetime.datetime(1890, 12, 31)]
        # get rid of people with wrong DateLeaving that is encoded as 1899-12-30 00:00:00 instead of dd.mm.yyyy
        df1 = df1[df1['DateLeaving'].str.len() == 10]
        # current members
        df2 = df[df['DateLeaving'].isnull()]
        # combine these dataframes
        df_after1890 = pd.concat([df1, df2])

        # generate unique ID for every person
        # generate two now columns
        df_after1890 = df_after1890.assign(uniqueIndex=0)
        df_after1890 = df_after1890.assign(DoubleName='')

        # group by first and last name, and date of birth
        grouped = df_after1890.groupby(["LastName", "FirstName", "DateOfBirth"])

        # assign first index to all entries of a person 
        for list_index in grouped.groups.values():
            df_after1890.loc[list_index, 'uniqueIndex'] = list_index[0]

        # some people are referred to by their double-name, add these double-namse to extra-column
        for row in self.df_exc.itertuples(index=False, name='Pandas'):
            df_after1890.loc[(df_after1890['LastName'] == row[0]) & (df_after1890['FirstName'] == row[1]), 'DoubleName'] = row[2]

#        print(df_after1890.loc[df_after1890['DoubleName'] != ''])
        # write dataframe to csv
        df_after1890.to_csv(self.output_file_csv)

        # for every year
        for year in self.range_years:

            # extract every MP that was active in that year
            # (every MP of a year joined before the end of the and left after the beginning of the year)
            df_year = df_after1890[pd.to_datetime(df_after1890['DateJoining']) <= datetime.datetime(year, 12, 31, 0, 0)]
            df_year = df_year[pd.to_datetime(df_year['DateLeaving']) >= datetime.datetime(year, 1, 1, 0, 0)]
            print(year, df_year.shape)
            
            # write df_year to a yearly csv file
        #    str_year = str(year)
        #    df_year.to_csv('home/lili/NLP_DemocraSci/nlp-democracy/output/MPs/MPs_' + str_year + '.csv')
            
            # create a pandas dataframe from list of names
            # !!! list contains errors, see definition of function
            list_lastnames = self.get_list_of_lastnames(df_year, df_after1890)
            df_lastnames = pd.DataFrame(list_lastnames, columns = ('type', 'name_short', 'name_correct', 'uniqueIndex', 'CantonName', 'CantonAbbreviation', 'Citizenship', 'FirstName'))

            # dump dictionary of last names to a pickle file
#           path = pathlib.
            with open(self.output_folder_dict + str(year) + "_lastnames.pickle", 'wb') as f:
                pickle.dump(df_lastnames, f)


# years of interest
years = [1891, 2016]   #2016

df_exc = pd.DataFrame(columns=['LastName', 'FirstName', 'DoubleName'])
# exception: Konrad H. Cramer is also reffered to as Cramer-Frey. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Cramer', 'FirstName': 'Konrad H.', 'DoubleName': 'Cramer-Frey'} 
# exception: Johannes Blumer SG is also reffered to as Blumer-Egloff. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Blumer', 'FirstName': 'Johannes', 'DoubleName': 'Blumer-Egloff'} 
# exception: Adolphe Jordan VD is also reffered to as Jordan-Martin. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Jordan', 'FirstName': 'Adolphe', 'DoubleName': 'Jordan-Martin'} 
# exception: Jakob Schmid LU is also reffered to as Schmid-Ronca. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Schmid', 'FirstName': 'Jakob', 'DoubleName': 'Schmid-Ronca'} 
# exception: Eduard Sulzer ZH is also reffered to as Sulzer-Ziegler. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Sulzer', 'FirstName': 'Eduard', 'DoubleName': 'Sulzer-Ziegler'} 
# exception: Howard Eugster AR is also reffered to as Eugster-Züst. Add double name in extra-column
df_exc.loc[len(df_exc)] = {'LastName': 'Eugster', 'FirstName': 'Howard', 'DoubleName': 'Eugster-Züst'} 
#print(df_exc)

mps_extractor = MPs_Extractor(years, input_file, output_file_csv, output_folder_dict, df_exc)
mps_extractor.extract()

